<?php
namespace APS;

class ExtendedReflectionClass extends \ReflectionClass {
	private $classSource = NULL;
	private $docComment = NULL;
	
	const MAX_DOC_COMMENT_OFFSET = 3;  
	const MAX_CLASS_DOC_COMMENT_OFFSET = 2;
	
	const DEFAULT_ARGUMENT_VALUE_OFFSET = 2; 
	
	const OPEN_BRACE = '{';
	const CLOSE_BRACE = '}';
	
	const COMMA = ',';	
	const EQUAL_SIGN = '=';
	
	const CLOSE_PARENTHESES = ')';
	
	function __construct($object) 
	{
		parent::__construct($object);	
		$this->parseClassSource();
	}
	
	public function getDocComment() 
	{
		return $this->docComment;
	}

	public function getConstructor() {
		$curr = $this;
		while(!is_null($curr)) 
		{
			$methods = $curr->getMethods();
			foreach($methods as $method) 
			{
				if($method->isConstructor()) 
				{
					return $method;
				}
			}
			$curr = new ExtendedReflectionClass(parent::getParentClass()->getName());
		}
		return NULL;
	}
	
	public function getProperties($filter = NULL) 
	{
		return $this->parseEntities(true, $filter);
	}
	
	public function getMethods($filter = NULL)
	{
		return $this->parseEntities(false, $filter);
	}
	
	public function getClassSource() 
	{
		return $this->classSource;
	}
 	
	protected function getAttributeNames($attributes) 
	{
		$names = array();
		foreach($attributes as $attribute) 
		{
			$names[] = $attribute->getName();
		}
		return $names;
	}
	
	protected function parseDocComment($classTokens, $tokenIndex) 
	{
		for($docIndex = 1; $docIndex <= self::MAX_DOC_COMMENT_OFFSET; $docIndex++)
		{
			if($classTokens[$tokenIndex - $docIndex][0] === \T_DOC_COMMENT)
			{
				return $classTokens[$tokenIndex - $docIndex][1];
			}
		}	
	}
	
	protected function isProperty($token, $beforeAttribute)
	{
		return $token[0] === \T_VARIABLE && 
		 		($beforeAttribute === \T_PUBLIC ||
				$beforeAttribute === \T_PROTECTED ||
				$beforeAttribute === \T_PRIVATE ||
				$beforeAttribute === \T_VAR ||
				$beforeAttribute === \T_STATIC); 
	}
	
	//TRUE - for parsing properties, FALSE - for parsing methods
 	protected function parseEntities($switch, $filter = NULL) 
 	{
 		$entities = array();
		$classTokens = $this->classSource;
		
		if(is_null($filter))
			if($switch) 
				$propertyNames = $this->getAttributeNames(parent::getProperties());
			else 
				$methodNames = $this->getAttributeNames(parent::getMethods());
		else
			if($switch)
				$propertyNames = $this->getAttributeNames(parent::getProperties($filter));
			else 
				$methodNames = $this->getAttributeNames(parent::getMethods($filter));
		
 		for($tokenIndex = 3; $tokenIndex < count($classTokens) - 1; $tokenIndex++) 
 		{
 			if(is_array($classTokens[$tokenIndex])) 
 			{
 				if($switch && $this->isProperty($classTokens[$tokenIndex], $classTokens[$tokenIndex - 1][0]) &&
 						in_array(substr($classTokens[$tokenIndex][1], 1), $propertyNames)) 
 				{
 					$entities[] = new ExtendedReflectionProperty(parent::getName(), 
 							substr($classTokens[$tokenIndex][1], 1), 
 							$this->parseDocComment($classTokens, $tokenIndex));
 				} 
 				elseif(!$switch && $classTokens[$tokenIndex][0] === \T_FUNCTION &&
 						in_array($classTokens[$tokenIndex + 1][1], $methodNames)) 
 				{
 					//T_FUNCTION +0 T_STRING +1 ( +2 firstArgument - +3
					$argumentString = $this->fetchMethodArgumentString($classTokens, $tokenIndex + 3);
					$methodParameters = $this->parseArguments($argumentString);
 					$entities[] = new ExtendedReflectionMethod(parent::getName(), $classTokens[$tokenIndex + 1][1],
 							$this->parseDocComment($classTokens, $tokenIndex), $methodParameters);
 				}
 			}
 		}
 		return $entities;
	}

//-------------------------------------------------------------------------
//Parse function arguments and create ReflectionParameter Objects
	
	protected function fetchMethodArgumentString($classTokens, $startIndex) 
	{
		$endIndex = $startIndex;
		while(strcmp($classTokens[$endIndex][0], self::CLOSE_PARENTHESES)) 
		{
			$endIndex++;
			if(!isset($classTokens[$endIndex])) {								
				throw new \ReflectionException("End of class exception");
			}
		}
		$argumentString = array_slice($classTokens, $startIndex, $endIndex - $startIndex);
		$this->removeCommasFromArgumentString($argumentString);
		return array_values($argumentString);
	}
	
	protected function removeCommasFromArgumentString(&$argumentString) 
	{
		foreach($argumentString as $index => $token) 
		{
			if(!strcmp($token[0], self::COMMA))
				unset($argumentString[$index]);
		}
	}
	
	protected function variableFilter($var ) 
	{
		return $var[0] === \T_VARIABLE;
	}
	
	protected function parseArguments($argumentString) 
	{
		$methodArguments = array(); 
		$indexes = array_keys(array_filter($argumentString, array($this, 'variableFilter')));
		foreach($indexes as $key => $argumentIndex) 
		{
			if(isset($argumentString[$argumentIndex + 1]) && 
				!strcmp($argumentString[$argumentIndex + 1][0], self::EQUAL_SIGN)) 
			{
				$methodArguments[] = $this->parseArgumentWithDefaultValues($argumentString, $argumentIndex, $key);
			} else {
				$methodArguments[] = $this->parseArgumentWithoutDefaultValues($argumentString, $argumentIndex, $key);
			}
		}
		return $methodArguments;
	}
	
	protected function parseArgumentWithDefaultValues($argumentString, $argumentIndex, $position) 
	{
		$defaultValue = $argumentString[$argumentIndex + self::DEFAULT_ARGUMENT_VALUE_OFFSET][1];
		$defaultValue = is_string($defaultValue) ? trim($defaultValue, "\"") : $defaultValue;
		return new ExtendedReflectionParameter(
						substr($argumentString[$argumentIndex][1], 1), 
						$position,
						true, 
						$defaultValue);		
	}
	
	protected function parseArgumentWithoutDefaultValues($argumentString, $argumentIndex, $position) 
	{
		return new ExtendedReflectionParameter(substr($argumentString[$argumentIndex][1], 1), $position);		
	}

//-------------------------------------------------------------------------
//Get Class Source for required Class

	protected function spaceFilter($var) 
	{
		return $var[0] != \T_WHITESPACE;
	}
		
	protected function parseClassSource() 
	{
		$tokens = token_get_all(file_get_contents(parent::getFileName()));
 		$tokensWithoutSpaces = array_values(array_filter($tokens, array($this, 'spaceFilter'))); 		
		$startIndex = $this->getClassStartIndex($tokensWithoutSpaces);
		$endIndex = $this->getClassEndIndex($tokensWithoutSpaces, $startIndex); 		
		$this->classSource = array_slice($tokensWithoutSpaces, $startIndex, $endIndex - $startIndex + 1);
	}
	
	protected function classFilter($var) 
	{
 		return $var[0] === \T_CLASS;
	}
	
	protected function getClassStartIndex($tokensWithoutSpaces) 
	{
		for($tokenIndex = 1; $tokenIndex < count($tokensWithoutSpaces) - 2; $tokenIndex++) 
		{
			if( isset($tokensWithoutSpaces[$tokenIndex][0]) && 
				$tokensWithoutSpaces[$tokenIndex][0] === \T_CLASS && 
				$tokensWithoutSpaces[$tokenIndex + 1][1] === end(explode('\\', parent::getName()))) 
			{
				for($docIndex = 1; $docIndex <= self::MAX_CLASS_DOC_COMMENT_OFFSET; $docIndex++) 
				{
					if(isset($tokensWithoutSpaces[$tokenIndex - $docIndex]) &&
						$tokensWithoutSpaces[$tokenIndex - $docIndex][0] === \T_DOC_COMMENT) 
					{
						$this->docComment = $tokensWithoutSpaces[$tokenIndex - $docIndex][1];
					}
				}
				return $tokenIndex;
			}
		}
		throw new \ReflectionException("Can't find class " . parent::getName() . " in source " . parent::getFileName());
	}
	
	protected function getClassEndIndex($tokenWithoutSpaces, $startLine) 
	{
		$braceCounter = 0;
		for($index = $startLine; isset($tokenWithoutSpaces[$index]); $index++) 
		{
			if(!strcmp($tokenWithoutSpaces[$index][0], self::OPEN_BRACE)) 
			{
				$braceCounter++;
			} 
			elseif (!strcmp($tokenWithoutSpaces[$index][0], self::CLOSE_BRACE)) 
			{
				$braceCounter--;
				if($braceCounter === 0) 
				{
					return $index;
				}
			}
		}
		throw new \ReflectionException("Not valid code: problem with braces.");
	}
}

class ExtendedReflectionProperty extends \ReflectionProperty
{
	private $docComment = NULL;

	function __construct($class, $name, $docComment = NULL)
	{
		parent::__construct($class, $name);
		$this->docComment = $docComment;
	}

	public function getDocComment()
	{
		return $this->docComment;
	}
}

class ExtendedReflectionMethod extends \ReflectionMethod
{

	private $docComment = NULL;
	private $parameters = NULL;

	function __construct($class, $name, $docComment = NULL, $parameters = NULL)
	{
		parent::__construct($class, $name);
		$this->docComment = $docComment;
		$this->parameters = $parameters;
	}

	public function getDocComment()
	{
		return $this->docComment;
	}

	public function getParameters()
	{
		return $this->parameters;
	}
}

class ReflectionParameter {}

class ExtendedReflectionParameter extends ReflectionParameter
{
	private $parameterName = NULL;
	private $position = 0;
	private $defaultValue = NULL;
	private $isOptional = false;

	function __construct($name, $position, $isOptional = false, $defaultValue = NULL)
	{
		$this->parameterName = $name;
		$this->position = $position;
		$this->isOptional = $isOptional;
		$this->defaultValue = $defaultValue;
	}

	public function getName()
	{
		return $this->parameterName;
	}

	public function getPosition()
	{
		return $this->position;
	}

	public function getDefaultValue()
	{
		return $this->defaultValue;
	}

	public function isOptional()
	{
		return $this->isOptional;
	}
}

?>