<?php
namespace Rest;

require_once 'aps/2/logger.php';
use APS;

/**
 * Interface for Rest service handlers strategies.
 */
interface StrategyInterface
{
	/**
	 * Returns request.
	 *
	 * @return RequestInterface RequestInterface instance
	 */
	public function getRequest();

	/**
	 * Returns response.
	 *
	 * @return ResponseInterface ResponseInterface instance
	 */
	public function getResponse();

	/**
	 * Prepares object before executing of requested method.
	 *
	 * @param mixed $object
	 * @param object $methodInfo
	 */
	public function prepareObject($object, $methodInfo);

	/**
	 * Executes post process.
	 *
	 * @param mixed $object
	 * @param object $methodInfo
	 * @param mixed $result
	 * @return mixed
	 */
	public function postProcess($object, $methodInfo, $result);
}

/**
 * Request objects interface.
 */
interface RequestInterface
{
	/**
	 *  Request method (GET|POST|PUT|DELETE|null).
	 *
	 * @return string
	 */
	public function getMethod();

	/**
	 * Returns request URI.
	 *
	 * @return string
	 */
	public function getUri();

	/**
	 * Returns request content.
	 *
	 * @return string
	 */
	public function getContent();

	/**
	 * Returns request HTTP Content-Type header value.
	 *
	 * @return string
	 */
	public function getContentType();

	/**
	 * Checks that request contains named parameter.
	 *
	 * @param string $name
	 * @return boolean
	 */
	public function hasParam($name);

	/**
	 * Returns request parameter.
	 *
	 * @param string $name
	 * @return string
	 */
	public function getParam($name);
}

/**
 * Response objects interface.
 */
interface ResponseInterface
{
	/**
	 * Adds HTTP header.
	 *
	 * @param string $name Header name
	 * @param string $value Header value
	 */
	public function addHeader($name, $value);

	/**
	 * Sets HTTP response code.
	 *
	 * @param integer $code
	 */
	public function setStatusCode($code, $message = null);

	/**
	 * Sets HTTP response content.
	 *
	 * @param string $content
	 */
	public function setContent($content);
}

/**
 * Default rest service strategy implementation.
 */
class DefaultServiceStrategy implements StrategyInterface
{
	/**
	 * Request object
	 * @var RequestInterface
	 */
	private $request;

	/**
	 * Response object
	 * @var ResponseInterface
	 */
	private $response;

	/**
	 * Strategy constructor
	 *
	 * @param RequestInterface $request
	 * @param ResponseInterface $response
	 */
	public function __construct(RequestInterface $request, ResponseInterface $response)
	{
		$this->request = $request;
		$this->response = $response;
	}

	public function getRequest()
	{
		return $this->request;
	}

	public function getResponse()
	{
		return $this->response;
	}

	public function prepareObject($object, $methodInfo)
	{
		$prepare = '_prepare'; // call fixed pre-process
		if(method_exists($methodInfo->class_name, $prepare)) {
			call_user_func_array(array($object, $prepare),
				array($methodInfo->method_name, &$methodInfo->arguments, $methodInfo));
		}

		if(method_exists($methodInfo->class_name, '_prepareData')) {
			call_user_func_array(array($object, '_prepareData'),
				array($methodInfo->method_name, $this->getRequest()->getContent()));
		}
	}

	public function postProcess($object, $methodInfo, $result)
	{
		$postprocess = '_postprocess'; // call fixed post-process
		if(method_exists($methodInfo->class_name, $postprocess)) {
			try {
				$result = call_user_func_array(array($object, $postprocess),
					array($methodInfo->method_name, $methodInfo->arguments, $result));
			} catch (NoContent $e) {
				$this->getResponse()->setStatusCode($e->getCode());
			}
		}

		return $result;
	}
}

class Request implements RequestInterface
{
	private $request_params = array();
	private $content;

	public function __construct()
	{
		$this->content = file_get_contents("php://input");
		if ($this->getMethod() == 'PUT' or $this->getMethod() == 'DELETE')
			parse_str($this->content, $this->request_params);
		if (isset($_POST))
			$this->request_params = array_merge($this->request_params, $_POST);
		if (isset($_GET))
			$this->request_params = array_merge($this->request_params, $_GET);
	}

	/**
	 * Returns server environment value if it exists
	 *
	 * @param string $key
	 * @return string|null
	 */
	private function _getFromServer($key)
	{
		return isset($_SERVER[$key])
			? $_SERVER[$key]
			: null;
	}

	public function getMethod()
	{
		return $this->_getFromServer('REQUEST_METHOD');
	}

	public function getUri()
	{
		return $this->_getFromServer('REQUEST_URI');
	}

	public function getContent()
	{
		return $this->content;
	}

	public function getContentType()
	{
		return $this->_getFromServer('CONTENT_TYPE');
	}

	public function hasParam($name)
	{
		return isset($this->request_params[$name]);
	}

	public function getParam($name)
	{
		if (isset($this->request_params[$name]))
			return $this->request_params[$name];
		return null;
	}
}

class Response implements ResponseInterface
{
	/**
	 * HTTP status codes
	 * @var array
	*/
	public static $codes = array(
		100 => 'Continue',
		101 => 'Switching Protocols',
		200 => 'OK',
		201 => 'Created',
		202 => 'Accepted',
		203 => 'Non-Authoritative Information',
		204 => 'No Content',
		205 => 'Reset Content',
		206 => 'Partial Content',
		300 => 'Multiple Choices',
		301 => 'Moved Permanently',
		302 => 'Found',
		303 => 'See Other',
		304 => 'Not Modified',
		305 => 'Use Proxy',
		306 => '(Unused)',
		307 => 'Temporary Redirect',
		400 => 'Bad Request',
		401 => 'Unauthorized',
		402 => 'Payment Required',
		403 => 'Forbidden',
		404 => 'Not Found',
		405 => 'Method Not Allowed',
		406 => 'Not Acceptable',
		407 => 'Proxy Authentication Required',
		408 => 'Request Timeout',
		409 => 'Conflict',
		410 => 'Gone',
		411 => 'Length Required',
		412 => 'Precondition Failed',
		413 => 'Request Entity Too Large',
		414 => 'Request-URI Too Long',
		415 => 'Unsupported Media Type',
		416 => 'Requested Range Not Satisfiable',
		417 => 'Expectation Failed',
		500 => 'Internal Server Error',
		501 => 'Not Implemented',
		502 => 'Bad Gateway',
		503 => 'Service Unavailable',
		504 => 'Gateway Timeout',
		505 => 'HTTP Version Not Supported'
	);

	public function addHeader($name, $value)
	{
		header($name . ': ' . $value);
	}

	public function setStatusCode($code, $message = null)
	{
		if (function_exists('http_response_code')) {
			\http_response_code($code);
		} else {
			$protocol = (isset($_SERVER['SERVER_PROTOCOL']) ? $_SERVER['SERVER_PROTOCOL'] : 'HTTP/1.1');
			header($protocol . ' ' . $code . ' ' . (is_null($message) ? static::$codes[strval($code)] : $message));
			$GLOBALS['http_response_code'] = $code;
		}
	}

	public function setContent($content)
	{
		echo $content;
	}
}

class RestService
{
	const VERSION = '2.0';

	/**
	 * Associated array that maps urls to their respective class and method
	 * @var array
	 */
	private $routes = array();

	/**
	 * Rest service handler strategy
	 * @var StrategyInterface
	 */
	private $strategy;

	/**
	 * Associated array that maps formats to their respective format class name
	 * @var array
	*/
	private $format_map = array();

	/**
	 * Service constructor
	 *
	 * @param array|null $routes List of routing rules
	 * @param StrategyInterface|null $strategy Rest service handler strategy
	 */
	public function __construct($routes = NULL, StrategyInterface $strategy = NULL)
	{
		$this->routes = $routes;

		if (is_null($strategy)) {
			$strategy = new DefaultServiceStrategy(new Request(), new Response());
		}
		$this->strategy = $strategy;
	}

	public function loadFormats($formats)
	{
		try {
			foreach($formats as $format)
			{
				$nameParts = explode('\\', $format);
				$formatName = end($nameParts);
				
				require_once 'aps/2/formats/'.lcfirst($formatName).'.php';
			}
		} catch (\Exception $e) {
			$this->handleError(500, 'File for '.$format.' format does not exist');
		}
	}

	/**
	 * Call this method and pass all the formats that should be supported by the API.
	 * @param string class name of the format class that implements iFormat
	 * @example $restService->setSupportedFormats('\Namespace\To\JsonFormat', '\Namespace\To\XmlFormat'...);
	 */
	public function setSupportedFormats() 
	{
		$args = func_get_args();
		if(empty($args))
			$this->handleError(500, 'There is no format specified in arguments of setSupportedFormats');
		$this->loadFormats($args);
		$this->format_map['default'] = $args[0];
		//No need to save first format, we save it as default and can shift off first element
		array_shift($args);
		$this->format_map['extensions'] = array();
		foreach ($args as $class_name) 
		{
			if(!class_exists($class_name))
				$this->handleError(500, $class_name.' does not exist');
			$obj = new $class_name;
			if(! $obj instanceof iFormat)
				$this->handleError(500, $class_name.' does not implement iFormat interface');
			foreach ($obj->getMIMEMap() as $extension => $mime) 
			{
				$this->format_map['extensions'][] = ".$extension";
				if(!isset($this->format_map[$mime]))
					$this->format_map[$mime] = $class_name;
			}
		}
	}
	
	public function getFormatAsObject($contentType)
	{
		if(isset($this->format_map[$contentType]))
			return new $this->format_map[$contentType];
		elseif(isset($this->format_map['default']))
			return new $this->format_map['default'] ($contentType);
		$this->handleError(500, 'No format class found for contentType'.$contentType);
	}
	
	public function getResponseFormatAsObject($contentType, $isSimpleResponse = false) 
	{
		if(is_null($contentType) && !$isSimpleResponse)
			$contentType = 'application/json';
		return $this->getFormatAsObject($contentType);
	}

	/**
	 * Convenience method to respond with an error message in case of Exception
	 * @param \Exception $e
	 */
	public function handleException($e)
	{ 
		$code = $e->getCode();
		if (!$code) 
			$code = 500;
		$class = get_class($e);
		$class = preg_replace('/\\\\/', '::', $class);
		
		$err = array(
			'code' => $code,
			'type' => $class,
			'message' => $e->getMessage()
		);

		if ($e->details)
			$err['details'] = $e->details;

		if (defined("APS_DEVELOPMENT_MODE"))
			$err['stacktrace'] = explode("\n", $e->getTraceAsString());
		
		$triggerError = null;
		if (defined("APS_DEVELOPMENT_MODE") && RestService::formatForHuman() && isset($e->xdebug_message))
		{
			// For human - send HTML representation of error
			$triggerError = $e->xdebug_message;
		}
			
		$this->sendError($err, $triggerError);
	}
	
	/**
	 * Convenience method to respond with an error message
	 * @param int $statusCode http error code
	 * @param string $errorMessage optional custom error message
	 * @param string $errorType optional error type (aka exception name)
	 */
	public function handleError($statusCode, $errorMessage = NULL) 
	{
		throw new \RuntimeException($errorMessage, $statusCode);
	}
	
	public function sendError($err, $triggerError = null)
	{
		$this->strategy->getResponse()->setStatusCode($err['code']);
		$fmt = $this->getResponseFormatAsObject('application/json');

		if ($err == null)
			throw new \Rest\RestException(500, "Unknown error");
			
		$errorText = $fmt->encode($err, defined("APS_DEVELOPMENT_MODE"));
		file_put_contents('php://stderr', "aps-runtime-error: " . $errorText);
		$this->sendData($err, $fmt, $triggerError);
	}
	

	/**
	 * Main function for processing the api request
	 * and return the response
	 * @throws \Exception when the api service class is missing
	 * @throws RestException to send error response
	 */
	public function handle() 
	{
        if (defined("APS_DEVELOPMENT_MODE")) {
            APS\Logger::get()->debug('Incoming Application API request: ' .
                $this->strategy->getRequest()->getMethod() . ' ' . $this->getRequestUrl());
        }

		if(empty($this->format_map))
			$this->handleError(500, 'No formats loaded');
		
		$o = $this->getAppropriateMethod();
		if(!isset($o->class_name))
			$this->handleError(404, 'No appropriate method found for url '.$this->getRequestUrl());
		
		try 
		{
			ob_start();
			
			$object = new $o->class_name();

			$result = null;
			
			$this->strategy->prepareObject($object, $o);

			$method = isset($o->override_method_name) ? $o->override_method_name : $o->method_name;
			if(method_exists($object, $method)) {
				try {
					$result = call_user_func_array(array($object, $method), $o->arguments);
				} catch (Accepted $e) {
					$this->strategy->getResponse()->setStatusCode($e->getCode(), $e->getMessage());
					$result = $e->response;
				}
			}

			$result = $this->strategy->postProcess($object, $o, $result);

			$fh = fopen("php://stderr", "a+");
			fwrite($fh, ob_get_contents());
			fclose($fh);
			ob_end_clean();
	    } catch (RestException $e) {
			$fh = fopen("php://stderr", "a+");
			$errorDetails = ob_get_contents();
			fwrite($fh, $errorDetails);
			fclose($fh);
			ob_end_clean();

			$e->details = $errorDetails;
			$this->handleException($e);
		}
		if (!is_null($result))
		{
			$isSimpleResponse = false;
			
			//This check for property existence is for old php 5.3.0, in 5.4.8 we can do like this
			//if(method_exists($o->response, 'isSimple'))
			//	$isSimpleResponse = $o->response->isSimple();
			//$contentType = $o->response->contentType;
			if(property_exists($o, 'response') && method_exists($o->response, 'isSimple'))
				$isSimpleResponse = $o->response->isSimple();
			$contentType = property_exists($o, 'response') ? $o->response->contentType : NULL;

			$this->sendData($result, $this->getResponseFormatAsObject($contentType, $isSimpleResponse));	
		}
	}
	
	public function sendData($data, $responseFormat, $triggerError = null)
	{
		$response = $this->strategy->getResponse();
		$response->addHeader('Cache-Control', 'no-cache, must-revalidate');
		$response->addHeader('Expires', '0');
		$response->addHeader('Content-Type', $triggerError ? "text/html" : $responseFormat->getMime());
		$response->addHeader('X-Powered-By', 'APS PHP Runtime '.self::VERSION);
	
		$data = $responseFormat->encode(
					$data, 
					defined("APS_DEVELOPMENT_MODE") || RestService::formatForHuman()
				);
				
		if ($triggerError)
		{
			$data = "<pre>" . $data . "</pre>";
			if (ini_get('html_errors')) 
			{
				$data .= "<br/><font size='1'>".
					"<table class='xdebug-error' dir='ltr' border='1' cellspacing='0' cellpadding='1'>".
					$triggerError.
					"</table>";
			}
		}
        if (defined("APS_DEVELOPMENT_MODE")) {
            APS\Logger::get()->debug("Result:\n$data");
        }
		$response->setContent($data);
	}
	
	public function getEncodedRequestData() {
		//get data from body, data as is - not decoded
		$encodedData = $this->strategy->getRequest()->getContent();
		if(is_null($encodedData))
			$this->handleError(500, 'Encoded data is null');
        if (defined("APS_DEVELOPMENT_MODE")) {
            APS\Logger::get()->debug("Request:\n$encodedData");
        }
		return $encodedData;
	}
			
	public function getRequestFormat() {
		$contentType = $this->strategy->getRequest()->getContentType();
		if(!is_null($contentType)) {
			//A media type is composed of two or more parts: A type, a subtype, and zero or more optional parameters.
			//Example - text/html; charset=UTF-8
			$contentTypeAsArray = explode(';', $contentType);
			return $contentTypeAsArray[0];
		}
		return NULL;
	}
	
	public function fillRequestDataByType($type) 
	{
		return $this->getFormatAsObject($this->getRequestFormat())->decode($this->getEncodedRequestData(), $type);
	}
	
	public function fillMethodParameters($method, $parametersViaPath = NULL) 
	{
		$method = (object)$method;
		foreach ($method->arguments as $name => $value)
			if (isset($parametersViaPath[$name]))
				$method->arguments[$name] = urldecode($parametersViaPath[$name]);
			elseif($this->strategy->getRequest()->hasParam($name))
				$method->arguments[$name] = $this->strategy->getRequest()->getParam($name);
			elseif(isset($method->arguments['request_data']))
			{
				$decodedData = $this->fillRequestDataByType($method->arguments['request_data']['type']);
				$method->arguments['request_data'] = $decodedData;
			}
			elseif(isset($method->defs[$name]))
				$method->arguments[$name] = $method->defs[$name];
		return $method;
	}
	
	public function getRequestUrl() {
		//Remove all slashes at the begin of string, at the end of string and request parameters(via ?)
		$requestUrl = preg_replace('/(^\/*)|(\/*\?.*$)|(\/*$)/', '', $this->strategy->getRequest()->getUri());
		//Remove extensions with dot in the end of request uri(for example, .xml or .json)
		return str_replace($this->format_map['extensions'], '', $requestUrl);
	}
	
	public function checkRouteUrl($routeUrl) 
	{
		$regex = preg_replace('/\\\{([^\}]+)\\\}/', '(?P<$1>[^/]+)', preg_quote($routeUrl));
		if (preg_match(":\b$regex$:", $this->getRequestUrl(), $matches))
			return $matches;
		return false;		
	}
	
	public function getRouteByRequestMethod($requestMethod)
	{
		if(!isset($this->routes[$requestMethod]))
			$this->handleError(500, 'No method found for corresponding url');
		return $this->routes[$requestMethod];
	}

	public function getAppropriateMethod() 
	{
		$routeUrls = $this->getRouteByRequestMethod($this->strategy->getRequest()->getMethod());
		foreach ($routeUrls as $routeUrl => $method)
		{
			if(strstr($routeUrl, '{') && ($parametersViaPath = $this->checkRouteUrl($routeUrl)))
				return $this->fillMethodParameters($method, $parametersViaPath);
			elseif (($this->getRequestUrl() === $routeUrl) ||
				    ($this->endsWith($this->getRequestUrl(), '/'.$routeUrl)))
				return $this->fillMethodParameters($method);
		}
	}

	public function endsWith($haystack, $needle)
	{
	    $length = strlen($needle);
    	if ($length == 0) {
	        return true;
	    }

    	return (substr($haystack, -$length) === $needle);
	}
	
	static public function formatForHuman()
	{
		if (!key_exists('SERVER_PROTOCOL', $_SERVER))
			return true; // command-line
		
		$human = false;		
		if (isset($_SERVER['HTTP_ACCEPT'])) 
		{
			$accept = $_SERVER['HTTP_ACCEPT'];
			$human = strstr($accept, "text/html") && !strstr($accept, "application/json");			
		}
		return $human;		
	}
}

class RestException extends \Exception
{
	public $details;

	public function __construct($http_status_code, $error_message = NULL, $details = NULL)
	{
		parent::__construct($error_message, $http_status_code);
		$this->details = $details;
	}
}

class Accepted extends \Exception
{
	public $response;

	public function __construct($response = NULL, $info = NULL, $timeout = NULL)
	{
		parent::__construct('Accepted', 202);
		
		if ($info)
			header("APS-Info: $info");

		if ($timeout)
			header("APS-Retry-Timeout: $timeout");

		$this->response = $response;
	}
}

class NoContent extends \Exception
{
    public $response;

    public function __construct()
    {
        parent::__construct('No Content', 204);
    }
}

interface iFormat
{
	/**
	 * Get Extension => MIME type mappings as an associative array
	 * @return array list of mime strings for the format
	 * @example array('json'=>'application/json');
	 */
	public function getMIMEMap();

	public function getMIME();

	public function encode($data);

	public function decode($data);
}
?>
